import pytest

from units_calculator.engine.engine import Number
from units_calculator.si_units.si_units import Meters, Seconds


def test_create_unitless_number() -> None:
    n = Number(5.0)
    assert repr(n) == "5.0"


def test_create_direct_unit() -> None:
    s = Seconds(5.0)
    assert repr(s) == "5.0s"


def test_compare_units() -> None:
    s1 = Seconds(2.0)
    s2 = Seconds(7.0)
    assert s1 < s2
    assert s1 <= s2
    assert s2 <= s2
    assert s2 == s2
    assert s2 >= s2
    assert s2 >= s1
    assert s2 > s1


def test_add_units() -> None:
    s1 = Seconds(4.0)
    s2 = Seconds(3)
    s3 = Seconds(5)
    _sum = s1 + s2
    assert repr(_sum) == "7.0s"
    _sum += s3
    assert _sum.base_units_val == 12.0


def test_sub_units() -> None:
    s1 = Seconds(12.0)
    s2 = Seconds(3)
    s3 = Seconds(5)
    _sum = s1 - s2
    assert repr(_sum) == "9.0s"
    _sum -= s3
    assert _sum.base_units_val == 4.0


def test_add_unit_and_number() -> None:
    s = Seconds(1.0)
    n = Number(-3)
    with pytest.raises(AssertionError):
        n += s  # type: ignore


def test_div_units() -> None:
    s1 = Seconds(7.0)
    s2 = Seconds(3.0)
    assert (s1 // s2).base_units_val == 2
    assert repr(s1 // s2) == "2.0"
    assert (s1 / s2).base_units_val == 7 / 3
    assert repr(s1 / s2) == repr(7 / 3)


def test_div_mixed() -> None:
    s1 = Seconds(7.0)
    assert (s1 / 2).base_units_val == 7.0 / 2
    m1 = Meters(3.0)
    assert repr(m1 / s1) == repr(3 / 7) + "m*s^(-1)"
    assert repr(s1 / m1) == repr(7 / 3) + "s*m^(-1)"
    assert repr(14 / s1) == "2.0s^(-1)"


def test_mod_units() -> None:
    s1 = Seconds(7.0)
    s2 = Seconds(5.0)
    assert repr(s1 % s2) == "2.0s"


def test_mul_units() -> None:
    s1 = Seconds(2.0)
    assert repr(s1 * s1) == "4.0s^2"
    m1 = Meters(7)
    assert repr(s1 * m1) == "14.0s*m"
    assert repr(m1 * s1) == "14.0m*s"
    assert s1 * m1 == m1 * s1
    assert repr(3 * m1) == "21.0m"


def test_pow_units() -> None:
    s1 = Seconds(2.0)
    n = Number(2)
    assert repr(s1 ** 2) == "4.0s^2"
    assert repr(s1 ** n) == "4.0s^2"
    assert repr(2 ** n) == "4.0"



"""A module for volume units"""

from units_calculator.derived_units.length_units import (
    Centimeters,
    Decimeters,
    Inches,
    Yards,
)
from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Meters


class CubicMeters(DerivedUnit):
    """Cubic meters"""

    __symbol__ = "m3"
    __base_units__ = [(Meters, 3)]


class CubicCentimeters(DerivedUnit):
    """Cubic centimeters"""

    __symbol__ = "cc"
    __base_units__ = [(Centimeters, 3)]


class Litre(DerivedUnit):
    """Litres"""

    __symbol__ = "l"
    __base_units__ = [(Decimeters, 3)]


class Decilitres(DerivedUnit):
    """Decilitres"""

    __symbol__ = "dl"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1e-1


class Centilitres(DerivedUnit):
    """Centilitres"""

    __symbol__ = "cl"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1e-2


class Millilitres(DerivedUnit):
    """Millilitres"""

    __symbol__ = "ml"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1e-3


class Microlitres(DerivedUnit):
    """Microlitres"""

    __symbol__ = "ul"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 1e-3


class Kilolitres(DerivedUnit):
    """Kilolitres"""

    __symbol__ = "kl"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1e3


class Megalitres(DerivedUnit):
    """Megalitres"""

    __symbol__ = "Ml"
    __base_units__ = [(Kilolitres, 1)]
    __multiplier__ = 1e3


class CubicInches(DerivedUnit):
    """Cubic inches"""

    __symbol__ = "in3"
    __base_units__ = [(Inches, 3)]


class CubicYard(DerivedUnit):
    """Cubic yard"""

    __symbol__ = "yd3"
    __base_units__ = [(Yards, 3)]


class ImperialFluidOunces(DerivedUnit):
    """Imperial fluid ounces"""

    __symbol__ = "imp_fl_oz"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 28.41306


class USFluidOunces(DerivedUnit):
    """US fluid ounces"""

    __symbol__ = "US_fl_oz"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 29.5735295625


class ImperialGallon(DerivedUnit):
    """Imperial gallons"""

    __symbol__ = "imp_gal"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 4.54609


class USGallons(DerivedUnit):
    """US gallons"""

    __symbol__ = "US_gal"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 3.785411784


class USDryGallons(DerivedUnit):
    """US dry gallons"""

    __symbol__ = "US_dry_gal"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 4.40488377086


class ImperialQuarts(DerivedUnit):
    """Imperial quarts"""

    __symbol__ = "imp_qt"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1.13652


class USQuarts(DerivedUnit):
    """US quarts"""

    __symbol__ = "US_qt"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 0.9463529460


class USDryQuarts(DerivedUnit):
    """US dry quarts"""

    __symbol__ = "US_dry_qt"
    __base_units__ = [(Litre, 1)]
    __multiplier__ = 1.101220942715


class ImperialPints(DerivedUnit):
    """Imperial pints"""

    __symbol__ = "imp_pt"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 568.26125


class USPints(DerivedUnit):
    """US pints"""

    __symbol__ = "US_pt"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 473.176473


class USDryPints(DerivedUnit):
    """US dry pints"""

    __symbol__ = "US_dry_pt"
    __base_units__ = [(Millilitres, 1)]
    __multiplier__ = 550.610471



"""A module containing metaclass for all unit classes"""

from __future__ import annotations

import copy
import math
from collections import defaultdict
from typing import Any, Optional, Type, Union, cast

from ordered_set import OrderedSet  # type: ignore

UNITS_CHARACTERS_BLACKLIST = "*/ "
UNITS_START_CHARACTERS_BLACKLIST = "0123456789.e+-^()j" + UNITS_CHARACTERS_BLACKLIST


class UnitsMeta(type):
    """A metaclass for all unit classes"""

    __acc_multiplier__: complex
    __dimensions__: Dimensions
    __symbol__: str

    def __new__(
        cls: Type[UnitsMeta],
        name: str,
        bases: tuple,  # type: ignore
        namespace: dict[str, Any],
    ) -> UnitsMeta:
        """Check required attributes exists in unit class"""
        # pylint: disable=too-many-locals
        symbol: Optional[str] = (
            namespace["__symbol__"] if "__symbol__" in namespace else None
        )
        if symbol is None:
            return cast(UnitsMeta, super().__new__(cls, name, bases, namespace))
        assert symbol[0] not in UNITS_START_CHARACTERS_BLACKLIST
        assert all(character not in UNITS_CHARACTERS_BLACKLIST for character in symbol)
        is_base_unit = symbol is not None and BaseUnit in bases
        is_derived_unit = symbol is not None and DerivedUnit in bases
        if symbol in dimensions_dict:
            raise RuntimeError(f"{symbol} defined twice!")
        if is_base_unit:
            idx = len(dimensions_dict)
            dimensions: Dimensions = list()
            namespace["__dimensions__"] = dimensions
            namespace["__acc_multiplier__"] = 1
            result = cast(UnitsMeta, super().__new__(cls, name, bases, namespace))
            dimensions.append((idx, 1, result))
            dimensions_dict[symbol] = idx, result
            idx_to_dimension[idx] = symbol, result
        elif is_derived_unit:
            src_dimensions: list[tuple[UnitsMeta, int]] = namespace["__base_units__"]
            dimensions = list()
            namespace["__dimensions__"] = dimensions
            multiplier: complex = namespace.get("__multiplier__", 1)
            dimensions_idx_dict: dict[int, Dimension] = dict()
            for src_dimension in src_dimensions:
                src_unit, derived_exp = src_dimension
                src_unit_dimensions: Dimensions = src_unit.__dimensions__
                for idx, source_exp, base_unit in src_unit_dimensions:
                    existing_dimension = dimensions_idx_dict.get(
                        idx, (idx, 0, base_unit)
                    )
                    existing_dimension = (
                        existing_dimension[0],
                        existing_dimension[1] + source_exp * derived_exp,
                        existing_dimension[2],
                    )
                    multiplier *= src_unit.__acc_multiplier__ ** derived_exp
                    dimensions_idx_dict[idx] = existing_dimension
            namespace["__acc_multiplier__"] = multiplier
            dimension_idx_keys = list(dimensions_idx_dict.keys())
            dimension_idx_keys.sort()
            for idx in dimension_idx_keys:
                _, exp, base_unit = dimensions_idx_dict[idx]
                if exp != 0:
                    dimensions.append((idx, exp, base_unit))
            result = cast(UnitsMeta, super().__new__(cls, name, bases, namespace))
            dimensions_dict[symbol] = None, result
        else:
            result = cast(UnitsMeta, super().__new__(cls, name, bases, namespace))
        return result


dimensions_dict: dict[str, tuple[Optional[int], UnitsMeta]] = dict()
idx_to_dimension: dict[int, tuple[str, UnitsMeta]] = dict()

# A list of dimensions and their exponents
Dimension = tuple[int, int, UnitsMeta]
Dimensions = list[tuple[int, int, UnitsMeta]]


class Unit(metaclass=UnitsMeta):
    """A base class for all units"""

    __dimensions__: Dimensions
    __symbol__: str
    __acc_multiplier__: complex

    def __init__(
        self,
        numerical_val: complex,
        dimensions: Dimensions,
        preferred_units: list[UnitsMeta],
    ):
        self._dimensions: Dimensions = dimensions
        self._preferred_units = preferred_units
        self._numerical_val: complex = numerical_val

    # Basic functionality
    def _is_matching_dimensions(self, other: Unit) -> bool:
        if len(self._dimensions) != len(
            other._dimensions  # pylint: disable=protected-access
        ):
            return False
        for dimension1, dimension2 in zip(
            self._dimensions, other._dimensions  # pylint: disable=protected-access
        ):
            idx1, exp1, _ = dimension1
            idx2, exp2, _ = dimension2
            if (idx1, exp1) != (idx2, exp2):
                return False
        return True

    def _clone(self) -> Unit:
        return Unit(
            self._numerical_val,
            copy.deepcopy(self._dimensions),
            copy.deepcopy(self._preferred_units),
        )

    @property
    def preferred_units_representation(self) -> list[tuple[UnitsMeta, int]]:
        """Returns preferred units for current unit dimensionality"""

        def _sum_dim_exp_squared(dimensions_exps: dict[int, int]) -> int:
            return sum(val ** 2 for val in dimensions_exps.values())

        def _add_n_unit_to_dimension(
            base_dimensions_exps: dict[int, int],
            unit_dimensions_exp: dict[int, int],
            n: int,
        ) -> dict[int, int]:
            return {
                idx: base_dimensions_exps[idx] + n * unit_dimensions_exp[idx]
                for idx in base_dimensions_exps.keys()
            }

        def _find_best_unit_exp(
            base_dimensions_exps: dict[int, int], unit_dimensions_exp: dict[int, int]
        ) -> int:
            current_sum_dim_exp_squared = _sum_dim_exp_squared(base_dimensions_exps)
            i = 0
            while True:
                next_unit_sum_dim_exp_squared = _sum_dim_exp_squared(
                    _add_n_unit_to_dimension(
                        base_dimensions_exps, unit_dimensions_exp, -i - 1
                    )
                )
                if next_unit_sum_dim_exp_squared >= current_sum_dim_exp_squared:
                    break
                i += 1
                current_sum_dim_exp_squared = next_unit_sum_dim_exp_squared
            if i > 0:
                return i
            i = 0
            while True:
                next_unit_sum_dim_exp_squared = _sum_dim_exp_squared(
                    _add_n_unit_to_dimension(
                        base_dimensions_exps, unit_dimensions_exp, -i + 1
                    )
                )
                if next_unit_sum_dim_exp_squared >= current_sum_dim_exp_squared:
                    break
                i -= 1
                current_sum_dim_exp_squared = next_unit_sum_dim_exp_squared
            return i

        result: list[tuple[UnitsMeta, int]] = list()
        remaining_dimensions: dict[int, int] = {
            idx: exp for idx, exp, _ in self._dimensions
        }
        for preferred_unit in self._preferred_units:
            preferred_unit_dimensions: dict[int, int] = {
                idx: exp for idx, exp, _ in preferred_unit.__dimensions__
            }
            for idx in remaining_dimensions.keys():
                if idx not in preferred_unit_dimensions:
                    preferred_unit_dimensions[idx] = 0
            for idx in preferred_unit_dimensions.keys():
                if idx not in remaining_dimensions:
                    remaining_dimensions[idx] = 0
            unit_exp = _find_best_unit_exp(
                remaining_dimensions, preferred_unit_dimensions
            )
            if unit_exp != 0:
                result.append((preferred_unit, unit_exp))
                remaining_dimensions = _add_n_unit_to_dimension(
                    remaining_dimensions, preferred_unit_dimensions, -unit_exp
                )
        idxs_to_pop: list[int] = list()
        for idx, exp in remaining_dimensions.items():
            if exp == 0:
                idxs_to_pop.append(idx)
        for idx in idxs_to_pop:
            remaining_dimensions.pop(idx)
        # Fill in remaining base dimensions
        for idx, _, _ in self._dimensions[::-1]:  # start with least common units
            if idx not in remaining_dimensions:
                continue
            result.append((idx_to_dimension[idx][1], remaining_dimensions[idx]))
        return result

    @property
    def _units_string_and_factor(self) -> tuple[str, complex]:
        """Return string representation of units"""
        units_representation_parts: list[str] = list()
        factor: complex = 1
        for unit, exp in self.preferred_units_representation:
            units_symbol: str = unit.__symbol__
            if exp != 1:
                units_symbol += f"^{str(exp) if exp > 0 else '(' + str(exp) + ')'}"
            units_representation_parts.append(units_symbol)
            factor *= unit.__acc_multiplier__ ** exp
        return "*".join(units_representation_parts), factor

    @property
    def representation_val(self) -> complex:
        """Return numerical value of unit it preferred representation"""
        return self._units_string_and_factor[1]

    @property
    def base_units_val(self) -> complex:
        """Return value in base units"""
        return self._numerical_val

    def _mrege_preferences(self, other: Unit) -> None:
        for (
            preferred_unit
        ) in other._preferred_units:  # pylint: disable=protected-access
            if preferred_unit not in self._preferred_units:
                self._preferred_units.append(preferred_unit)

    def __repr__(self) -> str:
        units_str, units_factor = self._units_string_and_factor
        val = self._numerical_val / units_factor
        numerical_representation = repr(val)
        units_representation = units_str
        return numerical_representation + units_representation

    # Arithmetic functionality
    def __lt__(self, other: Union[Unit, complex]) -> bool:
        if isinstance(other, Unit):
            assert self._is_matching_dimensions(other)
            return self._numerical_val < other._numerical_val  # type: ignore
        else:
            return self < Number(other)

    def __le__(self, other: Union[Unit, complex]) -> bool:
        if isinstance(other, Unit):
            assert self._is_matching_dimensions(other)
            return self._numerical_val <= other._numerical_val  # type: ignore
        else:
            return self <= Number(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Unit):
            assert self._is_matching_dimensions(other)
            return self._numerical_val == other._numerical_val
        else:
            if isinstance(other, (complex, float, int)):
                return self == Number(other)
            raise RuntimeError(f"Bad type in comparison with unit: {other}")

    def __ge__(self, other: Union[Unit, complex]) -> bool:
        if isinstance(other, Unit):
            assert self._is_matching_dimensions(other)
            return self._numerical_val >= other._numerical_val  # type: ignore
        else:
            return self >= Number(other)

    def __gt__(self, other: Union[Unit, complex]) -> bool:
        if isinstance(other, Unit):
            assert self._is_matching_dimensions(other)
            return self._numerical_val > other._numerical_val  # type: ignore
        else:
            return self > Number(other)

    def __iadd__(self, other: Unit) -> Unit:
        assert self._is_matching_dimensions(other)
        self._numerical_val += other._numerical_val
        return self

    def __add__(self, other: Unit) -> Unit:
        new_unit = self._clone()
        new_unit += other
        self._mrege_preferences(other)
        return new_unit

    def __neg__(self) -> Unit:
        self._numerical_val = -self._numerical_val
        return self

    def __sub__(self, other: Unit) -> Unit:
        return self + (-other)

    def __ifloordiv__(self, other: Unit) -> Unit:
        assert self._is_matching_dimensions(other)
        self._dimensions = list()
        self._numerical_val = math.floor(self._numerical_val / other._numerical_val)  # type: ignore
        self._mrege_preferences(other)
        return self

    def __floordiv__(self, other: Unit) -> Unit:
        new_unit = self._clone()
        new_unit //= other
        return new_unit

    def _itruediv_unit(self, other: Unit) -> Unit:
        it1, it2 = 0, 0
        result_dimensions: Dimensions = list()
        while it1 < len(
            self._dimensions  # pylint: disable=protected-access
        ) or it2 < len(
            other._dimensions  # pylint: disable=protected-access
        ):
            d1: Optional[Dimension] = None
            d2: Optional[Dimension] = None
            if it1 < len(self._dimensions):
                d1 = self._dimensions[it1]
            if it2 < len(other._dimensions):  # pylint: disable=protected-access
                d2 = other._dimensions[it2]  # pylint: disable=protected-access
            if d1 is not None and d2 is not None:
                if d1[0] == d2[0]:
                    if d1[1] != d2[1]:  # Remove dimensions with exponent 0
                        result_dimensions.append((d1[0], d1[1] - d2[1], d1[2]))
                    it1 += 1
                    it2 += 1
                elif d1[0] < d2[0]:
                    result_dimensions.append(d1)
                    it1 += 1
                else:
                    result_dimensions.append((d2[0], -d2[1], d2[2]))
                    it2 += 1
            elif d1 is not None:
                result_dimensions.append(d1)
                it1 += 1
            elif d2 is not None:
                result_dimensions.append((d2[0], -d2[1], d2[2]))
                it2 += 1
            else:
                raise RuntimeError("Bad dimensionality unit")
        self._dimensions = result_dimensions
        self._numerical_val /= other._numerical_val  # pylint: disable=protected-access
        self._mrege_preferences(other)
        return self

    def __itruediv__(self, other: Union[Unit, complex]) -> Unit:
        if isinstance(other, Unit):
            return self._itruediv_unit(other)
        else:
            self._numerical_val /= other
            return self

    def __truediv__(self, other: Union[Unit, complex]) -> Unit:
        new_unit = self._clone()
        new_unit /= other
        return new_unit

    def __rtruediv__(self, other: complex) -> Unit:
        result = Number(other) / self
        result._mrege_preferences(self)
        return result

    def __imod__(self, other: Unit) -> Unit:
        assert self._is_matching_dimensions(other)
        self._numerical_val %= other._numerical_val  # type: ignore
        self._mrege_preferences(other)
        return self

    def __mod__(self, other: Unit) -> Unit:
        new_unit = self._clone()
        new_unit %= other
        return new_unit

    def _imul_unit(self, other: Unit) -> Unit:
        it1, it2 = 0, 0
        result_dimensions: Dimensions = list()
        while it1 < len(
            self._dimensions
        ) or it2 < len(  # pylint: disable=protected-access
            other._dimensions  # pylint: disable=protected-access
        ):
            d1: Optional[Dimension] = None
            d2: Optional[Dimension] = None
            if it1 < len(self._dimensions):
                d1 = self._dimensions[it1]
            if it2 < len(other._dimensions):  # pylint: disable=protected-access
                d2 = other._dimensions[it2]  # pylint: disable=protected-access
            if d1 is not None and d2 is not None:
                if d1[0] == d2[0]:
                    if d1[1] != -d2[1]:  # Remove dimensions with exponent 0
                        result_dimensions.append((d1[0], d1[1] + d2[1], d1[2]))
                    it1 += 1
                    it2 += 1
                elif d1[0] < d2[0]:
                    result_dimensions.append(d1)
                    it1 += 1
                else:
                    result_dimensions.append(d2)
                    it2 += 1
            elif d1 is not None:
                result_dimensions.append(d1)
                it1 += 1
            elif d2 is not None:
                result_dimensions.append(d2)
                it2 += 1
            else:
                raise RuntimeError("Bad dimensionality unit")
        self._dimensions = result_dimensions
        self._numerical_val *= other._numerical_val  # pylint: disable=protected-access
        self._mrege_preferences(other)
        return self

    def __imul__(self, other: Union[Unit, complex]) -> Unit:
        if isinstance(other, Unit):
            return self._imul_unit(other)
        else:
            self._numerical_val *= other
            return self

    def __mul__(self, other: Union[Unit, complex]) -> Unit:
        new_unit = self._clone()
        new_unit *= other
        return new_unit

    def __rmul__(self, other: complex) -> Unit:
        result = Number(other) * self
        result._mrege_preferences(self)
        return result

    def _ipow_unit(self, other: Unit) -> Unit:
        # Exponent should be dimensionless
        assert len(other._dimensions) == 0  # pylint: disable=protected-access
        self **= other._numerical_val  # pylint: disable=protected-access
        self._mrege_preferences(other)
        return self

    def __ipow__(self, other: Union[Unit, complex]) -> Unit:
        if isinstance(other, Unit):
            return self._ipow_unit(other)
        else:
            if len(self._dimensions) > 0:
                if isinstance(other, complex):
                    exponent = other.real
                else:
                    exponent = other
                exponent = int(math.floor(exponent))
                assert exponent == other
            else:
                exponent = other  # type: ignore
            self._numerical_val **= exponent
            if len(self._dimensions) > 0:
                for i in range(len(self._dimensions)):
                    d: Dimension = self._dimensions[i]
                    d = (d[0], int(d[1] * exponent), d[2])
                    self._dimensions[i] = d
            return self

    def __pow__(self, other: Union[Unit, complex]) -> Unit:
        new_unit = self._clone()
        new_unit **= other
        return new_unit

    def __rpow__(self, other: complex) -> Unit:
        # Exponent must be dimensionless
        assert len(self._dimensions) == 0
        result = Number(other) ** self
        result._mrege_preferences(self)
        return result

    # Misc functions
    def as_base_units(self) -> Unit:
        """Return a unit without representation preferences"""
        new_unit = self._clone()
        new_unit._preferred_units = list()  # pylint: disable=protected-access
        return new_unit


class BaseUnit(Unit):
    """A class for base units"""

    def __init__(self, numerical_val: complex):
        super().__init__(numerical_val, self.__dimensions__, [self.__class__])


class DerivedUnit(Unit):
    """A base class for derived units"""

    __multiplier__: complex = 1
    __base_units__: list[tuple[UnitsMeta, int]]

    def __init__(self, numerical_val: complex):
        numerical_val *= self.__acc_multiplier__
        super().__init__(numerical_val, self.__dimensions__, [self.__class__])


class Number(Unit):
    """A class for unitless numbers"""

    __dimensions__: Dimensions = list()

    def __init__(self, numerical_val: complex):
        super().__init__(numerical_val, list(), list())


def parse_symbol(symbol: str) -> UnitsMeta:
    """Parse symbol as unit"""
    if symbol not in dimensions_dict:
        raise RuntimeError(f"Unrecognized symbol {symbol} in unit")
    return dimensions_dict[symbol][1]


def parse_pure_units(units_str: str) -> list[tuple[UnitsMeta, int]]:
    """Parse units string to list of units and exponents"""

    def _parse_pure_unit_exp(unit_exp_str: str) -> tuple[UnitsMeta, int]:
        unit_exp_parts = unit_exp_str.split("^")
        if len(unit_exp_parts) > 2:
            raise ValueError(f"Malformed unit: {unit_exp_str}")
        symbol = unit_exp_parts[0]
        unit = parse_symbol(symbol)
        exp = 1
        if len(unit_exp_parts) > 1:
            try:
                exp = int(unit_exp_parts[1].strip("()"))
            except ValueError as e:
                raise ValueError(
                    f"Cannot parse exponent {unit_exp_parts[1]} in unit {unit_exp_str}"
                ) from e
        return unit, exp

    def _parse_nodiv_unit_string(nodiv_str: str) -> list[tuple[UnitsMeta, int]]:
        _result: list[tuple[UnitsMeta, int]] = list()
        if len(nodiv_str) == 0:
            return _result
        units_exps = nodiv_str.split("*")
        for unit_exp in units_exps:
            _result.append(_parse_pure_unit_exp(unit_exp))
        return _result

    div_parts = units_str.split("/")
    if len(div_parts) > 2:
        raise ValueError("Cant have multiple '/' characters in unit string!")
    result_parts: list[tuple[UnitsMeta, int]] = _parse_nodiv_unit_string(div_parts[0])
    if len(div_parts) > 1:
        result_parts += [
            (unit, -exp) for (unit, exp) in _parse_nodiv_unit_string(div_parts[1])
        ]
    units = OrderedSet(unit for unit, _ in result_parts)
    result: list[tuple[UnitsMeta, int]] = list()
    for unit in units:
        exp = 0
        for res_unit, res_exp in result_parts:
            if res_unit is unit:
                exp += res_exp
        if exp != 0:
            result.append((unit, exp))
    return result


def _parse_numeric_val(numeric_part_str: str) -> complex:
    try:
        numeric_part: complex = int(numeric_part_str)
    except ValueError:
        try:
            numeric_part = float(numeric_part_str)
        except ValueError:
            try:
                numeric_part = complex(numeric_part_str)
            except ValueError as e:
                raise ValueError(
                    f"Cannot parse numerical value {numeric_part_str}"
                ) from e
    return numeric_part


def _units_to_dimensions(parsed_units: list[tuple[UnitsMeta, int]]) -> Dimensions:
    idx_to_exp: dict[int, int] = defaultdict(int)
    for unit, derived_exp in parsed_units:
        for idx, src_exp, _ in unit.__dimensions__:
            idx_to_exp[idx] += src_exp * derived_exp
    idxes = list(idx_to_exp.keys())
    idxes.sort()
    result: Dimensions = list()
    for idx in idxes:
        result.append((idx, idx_to_exp[idx], idx_to_dimension[idx][1]))
    return result


def parse(val_with_units: str) -> Unit:
    """Parse a string containing a units number+value into"""
    i = 0
    while val_with_units[i] in UNITS_START_CHARACTERS_BLACKLIST:
        i += 1
    numeric_part_str = val_with_units[:i]
    if len(numeric_part_str) == 0:
        raise ValueError(f"Cannot parse numerical part of unit{val_with_units}")
    numeric_part = _parse_numeric_val(numeric_part_str)
    parsed_units = parse_pure_units(val_with_units[i:])
    for unit, exp in parsed_units:
        numeric_part *= unit.__acc_multiplier__ ** exp
    return Unit(
        numeric_part,
        _units_to_dimensions(parsed_units),
        [unit for unit, exp in parsed_units],
    )



"""A module for chemistry units"""

from units_calculator.derived_units.volume_units import Litre
from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Mols


class Molars(DerivedUnit):
    """Molars"""

    __symbol__ = "M"
    __base_units__ = [(Mols, 1), (Litre, -1)]


class Millimolars(DerivedUnit):
    """Millimolars"""

    __symbol__ = "mM"
    __base_units__ = [(Molars, 1)]
    __multiplier__ = 1e-3



from units_calculator.derived_units.area_units import SquareCentimeters, SquareMeters
from units_calculator.derived_units.chemistry_units import Millimolars
from units_calculator.derived_units.mass_units import Grams, Milligrams
from units_calculator.derived_units.time_units import Milliseconds
from units_calculator.derived_units.volume_units import (
    CubicCentimeters,
    CubicInches,
    Millilitres,
    USQuarts,
)
from units_calculator.si_units.si_units import Meters, Mols, Seconds


def test_create_derived_unit() -> None:
    x = Milliseconds(4.0)
    assert repr(x) == "4.0ms"


def test_created_subderived_unit() -> None:
    x = Milligrams(6)
    assert repr(x) == "6.0mg"
    assert abs(x.base_units_val - 6e-6) < 1e-16


def test_add_mixed_units() -> None:
    s1 = Seconds(0.5)
    s2 = Milliseconds(100)
    assert repr(s1 + s2) == "0.6s"
    assert repr(s2 + s1) == "600.0ms"
    assert s1 > s2


def test_higher_dimensionality_units() -> None:
    a1 = SquareMeters(2)
    l1 = Meters(4)
    assert repr(a1) == "2.0m2"
    assert repr(a1 / l1) == "0.5m"


def test_high_dimensional_scaled_units() -> None:
    a1 = SquareCentimeters(50)
    assert repr(a1.as_base_units()) == "0.005m^2"


def test_multiple_pathways_to_same_unit() -> None:
    assert CubicCentimeters(5) == Millilitres(5)


def test_us_units() -> None:
    assert 57 < USQuarts(1) / CubicInches(1) < 58


def test_negative_dimensions_units() -> None:
    mass = Grams(10)
    molar_mass = Grams(294.31) / Mols(1)
    mols = mass / molar_mass
    assert Mols(0.033) < mols < Mols(0.034)
    volume = Millilitres(20)
    concentration = mols / volume
    assert 1.69 < (concentration / Millimolars(1)) < 1.7



"""A module containing weight unit definitions"""

from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Kilograms


class Grams(DerivedUnit):
    """Grams"""

    __symbol__ = "g"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 1e-3


class Milligrams(DerivedUnit):
    """Milligrams"""

    __symbol__ = "mg"
    __base_units__ = [(Grams, 1)]
    __multiplier__ = 1e-3


class Micrograms(DerivedUnit):
    """Micrograms"""

    __symbol__ = "ug"
    __base_units__ = [(Milligrams, 1)]
    __multiplier__ = 1e-3


class Nanograms(DerivedUnit):
    """Nanograms"""

    __symbol__ = "ng"
    __base_units__ = [(Milligrams, 1)]
    __multiplier__ = 1e-3


class Tonnes(DerivedUnit):
    """Metric tonne"""

    __symbol__ = "t"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 1e3


class Megatonnes(DerivedUnit):
    """Megatonnes"""

    __symbol__ = "Mt"
    __base_units__ = [(Tonnes, 1)]
    __multiplier__ = 1e6


class Gigatonnes(DerivedUnit):
    """Gigatonnes"""

    __symbol__ = "Gt"
    __base_units__ = [(Megatonnes, 1)]
    __multiplier__ = 1e3


class Daltons(DerivedUnit):
    """Daltons"""

    __symbol__ = "Da"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 1.66053906606050e-27


class Slugs(DerivedUnit):
    """Slugs"""

    __symbol__ = "sl"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 14.59390


class Pounds(DerivedUnit):
    """Pounds"""

    __symbol__ = "lb"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 0.45359237


class SolarMasses(DerivedUnit):
    """Solar masses"""

    __symbol__ = "M☉"
    __base_units__ = [(Kilograms, 1)]
    __multiplier__ = 1.98847e30



"""A module for SI units definition.
See https://en.wikipedia.org/wiki/International_System_of_Units"""

from units_calculator.engine.engine import BaseUnit


class Seconds(BaseUnit):
    """Seconds"""

    __symbol__: str = "s"


class Meters(BaseUnit):
    """Meters"""

    __symbol__: str = "m"


class Kilograms(BaseUnit):
    """Kilograms"""

    __symbol__: str = "kg"


class Amperes(BaseUnit):
    """Amperes"""

    __symbol__: str = "A"


class Kelvins(BaseUnit):
    """Kelvins"""

    __symbol__: str = "K"


class Mols(BaseUnit):
    """Mols"""

    __symbol__: str = "mol"


class Candelas(BaseUnit):
    """Candelas"""

    __symbol__: str = "cd"



"""A module containing common derived time units"""

from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Seconds


class Milliseconds(DerivedUnit):
    """Milliseconds"""

    __symbol__ = "ms"
    __base_units__ = [(Seconds, 1)]
    __multiplier__ = 1e-3


class Microseconds(DerivedUnit):
    """Microseconds"""

    __symbol__ = "us"
    __base_units__ = [(Milliseconds, 1)]
    __multiplier__ = 1e-3


class Nanoseconds(DerivedUnit):
    """Nanoseconds"""

    __symbol__ = "ns"
    __base_units__ = [(Microseconds, 1)]
    __multiplier__ = 1e-3


class Gigaseconds(DerivedUnit):
    """Gigaseconds"""

    __symbol__ = "Gs"
    __base_units__ = [(Seconds, 1)]
    __multiplier__ = 1e9


class Teraseconds(DerivedUnit):
    """Teraseconds"""

    __symbol__ = "Ts"
    __base_units__ = [(Gigaseconds, 1)]
    __multiplier__ = 1e3


class Minutes(DerivedUnit):
    """Minutes"""

    __symbol__ = "min"
    __base_units__ = [(Seconds, 1)]
    __multiplier__ = 60


class Hours(DerivedUnit):
    """Hours"""

    __symbol__ = "h"
    __base_units__ = [(Minutes, 1)]
    __multiplier__ = 60


class Days(DerivedUnit):
    """Days"""

    __symbol__ = "d"
    __base_units__ = [(Hours, 1)]
    __multiplier__ = 24


class Weeks(DerivedUnit):
    """Weeks"""

    __symbol__ = "weeks"
    __base_units__ = [(Days, 1)]
    __multiplier__ = 7


class JulianYear(DerivedUnit):
    """Julian years"""

    __symbol__ = "a"
    __base_units__ = [(Days, 1)]
    __multiplier__ = 365.25



"""A wrapper module to import whole library"""

# pylint: disable=unused-import

from .derived_units.area_units import (
    Acres,
    Barn,
    Hectares,
    SquareCentimeters,
    SquareFeet,
    SquareKilometers,
    SquareMeters,
    SquareMiles,
    SquareMillimeters,
    SquareYards,
)
from .derived_units.length_units import (
    Angstroms,
    AstronomicalUnits,
    Centimeters,
    Decimeters,
    EarthRadii,
    Feet,
    Inches,
    Kilometers,
    LightYears,
    LunarDistances,
    Micrometers,
    Miles,
    Millimeters,
    Nanometers,
    NauticalMiles,
    Parsec,
    Picometers,
    Yards,
)
from .derived_units.mass_units import (
    Daltons,
    Gigatonnes,
    Grams,
    Megatonnes,
    Micrograms,
    Milligrams,
    Nanograms,
    Pounds,
    Slugs,
    SolarMasses,
    Tonnes,
)
from .derived_units.time_units import (
    Days,
    Gigaseconds,
    Hours,
    JulianYear,
    Microseconds,
    Milliseconds,
    Minutes,
    Nanoseconds,
    Teraseconds,
    Weeks,
)
from .derived_units.volume_units import (
    Centilitres,
    CubicCentimeters,
    CubicInches,
    CubicMeters,
    CubicYard,
    Decilitres,
    ImperialFluidOunces,
    ImperialGallon,
    ImperialPints,
    ImperialQuarts,
    Kilolitres,
    Litre,
    Megalitres,
    Microlitres,
    Millilitres,
    USDryGallons,
    USDryPints,
    USDryQuarts,
    USFluidOunces,
    USGallons,
    USPints,
    USQuarts,
)
from .engine.engine import (
    BaseUnit,
    DerivedUnit,
    Number,
    Unit,
    UnitsMeta,
    parse,
    parse_pure_units,
    parse_symbol,
)
from .si_units.si_units import (
    Amperes,
    Candelas,
    Kelvins,
    Kilograms,
    Meters,
    Mols,
    Seconds,
)



"""A module containing area units definitions"""

from units_calculator.derived_units.length_units import (
    Centimeters,
    Feet,
    Kilometers,
    Miles,
    Millimeters,
    Yards,
)
from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Meters


class SquareMeters(DerivedUnit):
    """Square meters"""

    __symbol__ = "m2"
    __base_units__ = [(Meters, 2)]


class SquareCentimeters(DerivedUnit):
    """Square centimeters"""

    __symbol__ = "cm2"
    __base_units__ = [(Centimeters, 2)]


class SquareMillimeters(DerivedUnit):
    """Square millimeters"""

    __symbol__ = "mm2"
    __base_units__ = [(Millimeters, 2)]


class SquareKilometers(DerivedUnit):
    """Square kilometers"""

    __symbol__ = "km2"
    __base_units__ = [(Kilometers, 2)]


class SquareFeet(DerivedUnit):
    """Square feet"""

    __symbol__ = "ft2"
    __base_units__ = [(Feet, 2)]


class SquareYards(DerivedUnit):
    """Square yards"""

    __symbol__ = "yd2"
    __base_units__ = [(Yards, 2)]


class SquareMiles(DerivedUnit):
    """Square miles"""

    __symbol__ = "mi2"
    __base_units__ = [(Miles, 2)]


class Acres(DerivedUnit):
    """Acres"""

    __symbol__ = "ac"
    __base_units__ = [(Meters, 2)]
    __multiplier__ = 4064.9


class Hectares(DerivedUnit):
    """Hectares"""

    __symbol__ = "ha"
    __base_units__ = [(Meters, 2)]
    __multiplier__ = 1e4


class Barn(DerivedUnit):
    """Barns"""

    __symbol__ = "b"
    __base_units__ = [(Meters, 2)]
    __multiplier__ = 1e-28



"""Package setup file"""

import pathlib

from setuptools import find_packages, setup  # type: ignore

__version__ = "1.0.0"

packages = find_packages(exclude=["tests"])

HERE = pathlib.Path(__file__).parent
README_CONTENT = (HERE / "README.md").read_text()

setup(
    name="units-calculator",
    version=__version__,
    description="A units calculator for dimensional analysis,"
    " optimized for comfort over performance",
    long_description=README_CONTENT,
    long_description_content_type="text/markdown",
    url="https://github.com/KanHarI/units-calculator",
    author="Itay KanHar",
    author_email="knaan.harpaz@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    package_data={package: ["py.typed"] for package in packages},
    packages=find_packages(),
    install_requires=["ordered-set"],
)



from units_calculator.all import (
    Kilograms,
    Meters,
    Milliseconds,
    Seconds,
    parse,
    parse_pure_units,
)


def test_parse_pure_units() -> None:
    test_str = "s^(-1)*kg*m/s"
    x = parse_pure_units(test_str)
    assert tuple(x) == ((Seconds, -2), (Kilograms, 1), (Meters, 1))
    assert tuple(parse_pure_units("")) == ()


def test_prase_int_units() -> None:
    _5m = parse("5m")
    _25m2 = parse("25m^2")
    assert repr(_5m) == "5.0m"
    assert _5m * _5m == _25m2


def test_parse_vloat_units() -> None:
    _2500ms = parse("2.5s")
    assert _2500ms == Milliseconds(2500)


def test_parse_complex_units() -> None:
    _jA = parse("jA")
    assert repr(_jA) == "1jA"
    _3p2jA = parse("(3+2.5j)A")
    assert repr(_3p2jA) == "(3+2.5j)A"



"""A module containing distance units definitions"""

from units_calculator.engine.engine import DerivedUnit
from units_calculator.si_units.si_units import Meters


class Decimeters(DerivedUnit):
    """Decimeters"""

    __symbol__ = "dm"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1e-1


class Centimeters(DerivedUnit):
    """Centimeters"""

    __symbol__ = "cm"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1e-2


class Millimeters(DerivedUnit):
    """Millimeters"""

    __symbol__ = "mm"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1e-3


class Micrometers(DerivedUnit):
    """Micrometers"""

    __symbol__ = "um"
    __base_units__ = [(Millimeters, 1)]
    __multiplier__ = 1e-3


class Nanometers(DerivedUnit):
    """Nanometers"""

    __symbol__ = "nm"
    __base_units__ = [(Micrometers, 1)]
    __multiplier__ = 1e-3


class Angstroms(DerivedUnit):
    """Angstroms"""

    __symbol__ = "Å"
    __base_units__ = [(Nanometers, 1)]
    __multiplier__ = 1e-10


class Picometers(DerivedUnit):
    """Picometers"""

    __symbol__ = "pm"
    __base_units__ = [(Nanometers, 1)]
    __multiplier__ = 1e-3


class Kilometers(DerivedUnit):
    """Kilometers"""

    __symbol__ = "km"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1e3


class Inches(DerivedUnit):
    """Inches"""

    __symbol__ = "in"
    __base_units__ = [(Millimeters, 1)]
    __multiplier__ = 25.4


class Feet(DerivedUnit):
    """Feet"""

    __symbol__ = "ft"
    __base_units__ = [(Inches, 1)]
    __multiplier__ = 12


class Yards(DerivedUnit):
    """Yards"""

    __symbol__ = "yd"
    __base_units__ = [(Feet, 1)]
    __multiplier__ = 3


class Miles(DerivedUnit):
    """Terrestrial miles"""

    __symbol__ = "mi"
    __base_units__ = [(Yards, 1)]
    __multiplier__ = 1760


class NauticalMiles(DerivedUnit):
    """Nautical miles"""

    __symbol__ = "nmi"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1852


class EarthRadii(DerivedUnit):
    """Earth radii"""

    __symbol__ = "R⊕"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 6.3781e6


class LunarDistances(DerivedUnit):
    """Lunar distances"""

    __symbol__ = "LD"
    __base_units__ = [(Kilometers, 1)]
    __multiplier__ = 384402


class AstronomicalUnits(DerivedUnit):
    """Astronomical units"""

    __symbol__ = "au"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 1.495978707e11


class LightYears(DerivedUnit):
    """Light years"""

    __symbol__ = "ly"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 9.4607e15


class Parsec(DerivedUnit):
    """Parsecs"""

    __symbol__ = "pc"
    __base_units__ = [(Meters, 1)]
    __multiplier__ = 3.0857e16
